#!/usr/bin/env perl
# -*- mode: Perl; tab-width: 2; indent-tabs-mode: nil -*- # For Emacs

# CSV Plot
#
# Author(s): Martial Michel
#
# This software was developed at the National Institute of Standards and Technology by
# employees and/or contractors of the Federal Government in the course of their official duties.
# Pursuant to Title 17 Section 105 of the United States Code this software is not subject to 
# copyright protection within the United States and is in the public domain.
#
# "CSV Plot" is an experimental system.
# NIST assumes no responsibility whatsoever for its use by any party.
#
# THIS SOFTWARE IS PROVIDED "AS IS."  With regard to this software, NIST MAKES NO EXPRESS
# OR IMPLIED WARRANTY AS TO ANY MATTER WHATSOEVER, INCLUDING MERCHANTABILITY,
# OR FITNESS FOR A PARTICULAR PURPOSE.

use strict;

# Note: Designed for UNIX style environments (ie use cygwin under Windows).

##########
# Version

# $Id: 2factor_csv_plot.pl,v 1.2 2010/09/22 21:53:35 martial Exp $
my $version     = "0.1b";

if ($version =~ m/b$/) {
  (my $cvs_version = '$Revision: 1.2 $') =~ s/[^\d\.]//g;
  $version = "$version (CVS: $cvs_version)";
}

my $versionid = "CSV Plot Version: $version";

##########
# Check we have every module (perl wise)

my ($f4b, @f4bv);
BEGIN {
  $f4b = "F4DE_BASE";
  push @f4bv, (exists $ENV{$f4b}) 
    ? ($ENV{$f4b} . "/lib") 
      : ("../../../common/lib");
}
use lib (@f4bv);

sub eo2pe {
  my $oe = join(" ", @_);
  return( ($oe !~ m%^Can\'t\s+locate%) ? "\n----- Original Error:\n $oe\n-----" : "");
}

## Then try to load everything
my $have_everything = 1;
my $partofthistool = "It should have been part of this tools' files. Please check your $f4b environment variable.";
my $warn_msg = "";
sub _warn_add { $warn_msg .= "[Warning] " . join(" ", @_) ."\n"; }

# Part of this tool
foreach my $pn ("MMisc", "CSVHelper") {
  unless (eval "use $pn; 1") {
    my $pe = &eo2pe($@);
    &_warn_add("\"$pn\" is not available in your Perl installation. ", $partofthistool, $pe);
    $have_everything = 0;
  }
}

# usualy part of the Perl Core
foreach my $pn ("Getopt::Long") {
  unless (eval "use $pn; 1") {
    &_warn_add("\"$pn\" is not available on your Perl installation. ", "Please look it up on CPAN [http://search.cpan.org/]\n");
    $have_everything = 0;
  }
}

# Something missing ? Abort
if (! $have_everything) {
  print "\n$warn_msg\nERROR: Some Perl Modules are missing, aborting\n";
  exit(1);
}

# Use the long mode of Getopt
Getopt::Long::Configure(qw(auto_abbrev no_ignore_case));

########################################
# Options processing

my $usage = &set_usage();
MMisc::error_quit($usage) 
  if (scalar @ARGV == 0);

# Default values for variables
# none here

my $mainlabel = "";
my $xlabel = "";
my $ylabel = "";
my $rlf = 0;
my $fflm = 0;
my $fflm_plot = 0;

# Av  : ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz  #
# Used:      F           r             f h   l         v xy   #

my %opt = ();
GetOptions
  (
   \%opt,
   'help',
   'version',
   'label=s'  => \$mainlabel,
   'xlabel=s' => \$xlabel,
   'ylabel=s' => \$ylabel,
   'robustLineFitting' => \$rlf,
   'fitterFunctionsforLinearModels' => \$fflm,
   'FitterPlot' => \$fflm_plot,
  ) or MMisc::error_quit("Wrong option(s) on the command line, aborting\n\n$usage\n");
MMisc::ok_quit("\n$usage\n") if (($opt{'help'}) || (scalar @ARGV == 0));
MMisc::ok_quit("$versionid\n") if ($opt{'version'});

MMisc::ok_quit($usage) if (scalar @ARGV != 4);

my ($csv_file, $col1, $col2, $plot_file) = @ARGV;

my $tool = MMisc::cmd_which('R');
MMisc::error_quit("Could not find the \'R\' executable in your PATH")
  if (! defined $tool);
my $R = "$tool --vanilla --quiet --slave";

&check_csv_columns($csv_file, $col1, $col2);
&do_plot("commands.R", $plot_file, $csv_file, $col1, $col2);

MMisc::ok_quit("Done");

####################

sub csverr {
  MMisc::error_quit("No CSV handler ?") if (! defined $_[0]);
  MMisc::error_quit("CSV handler problem: " . $_[0]->get_errormsg())
      if ($_[0]->error());
}

##

sub get_csv_file_columns {
  my ($csvf) = @_;

  my $csvh = new CSVHelper();
  &csverr($csvh);
  my @h = $csvh->loadCSV_getheader($csvf);
  &csverr($csvh);

  return(@h);
}

#####

sub check_csv_columns {
  my ($csv_file, $col1, $col2) = @_;

  print "[*] Checking CSV file columns\n";

  my $err = MMisc::check_file_r($csv_file);
  MMisc::error_quit("Problem with \'csvfile\' ($csv_file) : $err")
      if (! MMisc::is_blank($err));

  MMisc::error_quit("Columns identical ? ($col1)")
      if ($col1 eq $col2);

  my @cols = &get_csv_file_columns($csv_file);
  my %tmp = MMisc::array1d_to_count_hash(\@cols);
  my $errtxt = "";
  foreach my $c ($col1, $col2) {
    if (! MMisc::safe_exists(\%tmp, $c)) {
      $errtxt .= " - Required column [$c] not found in CSV file [$csv_file]\n";
      next;
    }
    if ($tmp{$c} > 1) {
      $errtxt .= " - Required column [$c] has multiple candidates (" 
        . $tmp{$c} . ") in CSV file [$csv_file]\n";
    }
  }
  MMisc::error_quit("Problem with provided CSV file [$csv_file] columns :\n $errtxt")
      if (! MMisc::is_blank($errtxt));
}

##########

sub do_plot {
  my ($rfile, $png_file, $csv_file, $col1, $col2) = @_;

  my $png = MMisc::get_file_full_path($png_file);
  my $csv = MMisc::get_file_full_path($csv_file);

  my $Rcmds = "";
  
  $Rcmds .= "csv <- read.csv(file=\"$csv\", head=TRUE, sep=\",\") \n";

  $Rcmds .= "png(file=\'$png\') \n";

  $Rcmds .= "plot(csv\$$col1, csv\$$col2";
  $Rcmds .= (MMisc::is_blank($mainlabel)) ? "" : ", main=\"$mainlabel\"";
  $Rcmds .= (MMisc::is_blank($xlabel)) ? "" : ", xlabel=\"$xlabel\"";
  $Rcmds .= (MMisc::is_blank($ylabel)) ? "" : ", ylabel=\"$ylabel\"";
  $Rcmds .= ") \n";

  # Robust Line Fitting
  $Rcmds .= (! $rlf) ?  "" :
    ( "z <- line(csv\$$col1, csv\$$col2) \n" 
    . "abline(coef(z)) \n"
    . "print(\"[Robust Line Fitting] y=Ax+B\") \n"
    . "print(coef(z)) \n" );
  
  # Fitter Functions for Linear Models
  $Rcmds .= (! $fflm) ? "" :
    ( "fit <- lm(formula = csv\$$col1 ~ csv\$$col2) \n"
    . "print(\"[Fitter Functions for Linear Models]\") \n"
    . "summary(fit) \n" )
    . (($fflm_plot) ? "plot(fit) \n" : "");  

  $Rcmds .= "print(\"R Done\")\n\n";

  MMisc::error_quit("Problem creating R command file ($rfile)")
    if (! MMisc::writeTo($rfile, "", 1, 0, $Rcmds));

  my $log = "$rfile.log";

  my ($ok, $otxt, $so, $se, $rc, $log) = MMisc::write_syscall_logfile($log, "$R < $rfile");
  MMisc::error_quit("Problem running R ($R) using command file ($rfile), see $log")
    if ($rc != 0);

  print "[R printout]\n$so\n";
}

##########

sub set_usage {
  return("$0 infile.csv col1name col2name plotfile.png");
}
